//============================================================================
//  Copyright (c) 2000 University of Kansas 
//  All rights reserved.
//
//  KU MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF 
//  THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
//  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
//  PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  KU SHALL NOT BE LIABLE
//  FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING,
//  MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
//
//
//  By using or copying this Software, Licensee agrees to abide by the
//  intellectual property laws, and all other applicable laws of the
//  U.S., and the terms of this license.
//
//	authors: Cindy ckong@ittc.ukans.edu
//
//  $Id: SOM_FunctionType.java,v 1.1 2000/07/11 00:11:11 cvs Exp $
//
//  Revision History:
//  ----------------
//  $Log: SOM_FunctionType.java,v $
//  Revision 1.1  2000/07/11 00:11:11  cvs
//  Change in the grammar: type, set, bunch, array, record, tuple and sequence are
//  keywords. Typechecking for set is done. Added SetType and FunctionType.
//  False and true are predefined as boolean.
//  -- Cindy
//
//
//=============================================================================
package SOM;

import util.ListEnumeration;

/** This represents the <B>FunctionType</B> object
 * @author <I>Cindy Kong</I> (ckong@ittc.ukans.edu)
 */
public class SOM_FunctionType
extends SOM_TypeDeclaration {

	// to save the domains of the function type
	protected SOM_DeclarationList domainList = new SOM_DeclarationList();

	// This will be the range of this function type
	protected SOM_TypeDeclaration range;

	// This string is used simply to keep consistent with type
	// declaration having a name
	static private SOM_Identifier iden = new SOM_Identifier("functionType");

	/** The primary constructor
	 * @param myType The SOM_TypeDeclaration corresponding to the
	 * "bunch MetaType"
	 * @return Not applicable
	 */
	public SOM_FunctionType(SOM_TypeDeclaration myType) {
		super(iden, myType);
	}

	/** To add to the domainList
	 * @param domainIn The SOM_Declaration corresponding to (one of)
	 * the domains of this operator
	 * @return void
	 */
	public void addToDomainList(SOM_Declaration domainIn) {
		domainList.addElement(domainIn);
	}

	/** To get the domainList
	 * @return The SOM_DeclarationList of the domains for this function
	 */
	public SOM_DeclarationList getDomainList() {
		return domainList;
	}

	/** To set the range
	 * @param rangeIn The SOM_TypeDeclaration corresponding to the
	 * range of this operator
	 * @return void
	 */
	public void setRange(SOM_TypeDeclaration rangeIn) {
		range = rangeIn;
	}

	/** To get the return type
	 * @return The SOM_TypeDelcaration corresponding to the range - 
	 * the return type
	 */
	public SOM_TypeDeclaration getRange() {
		return range;
	}

	/** To implement a version of isDistinguishable for function types
	 * @param typeOne The first SOM_FunctionType
	 * @param typeTwo The next SOM_FunctionType to be distinguished
	 * against typeOne
	 * @return true if they are distinguishable, false otherwise
	 */
	public static boolean isDistinguishable(SOM_FunctionType typeOne,
											SOM_FunctionType typeTwo) { 
		SOM_DeclarationList typeListOne = typeOne.getDomainList();
		SOM_DeclarationList typeListTwo = typeTwo.getDomainList();

		if (typeListOne.countElements() != typeListTwo.countElements())
			return true;

		ListEnumeration enumOne = typeListOne.elements();
		ListEnumeration enumTwo = typeListTwo.elements();

		for (int i = 0; i < typeListOne.countElements(); i++) {
			if (SOM_TypeDeclaration.isDistinguishable
					((SOM_TypeDeclaration)enumOne.nextElement(),
					 (SOM_TypeDeclaration)enumTwo.nextElement())
				 )
				return true;
		}

		return SOM_TypeDeclaration.isDistinguishable
			(typeOne.getRange(), typeTwo.getRange());
	}

	/** To ascertain "compatiblity" for bindings of the kind lhs <- rhs
	 * @param typeLHS The FunctionType corresponding to the lhs object
	 * @param typeRHS The FunctionType corresponding to the rhs object
	 * @return true, if the rhs is assignable to the lhs, false otherwise
	 */
	public static boolean
			isLeftCompatibleWithRight(SOM_FunctionType typeLHS,
									  SOM_FunctionType typeRHS) { 
		SOM_DeclarationList domainListLHS = typeLHS.getDomainList();
		SOM_DeclarationList domainListRHS = typeRHS.getDomainList();

		if (domainListLHS.countElements() != domainListRHS.countElements())
			// if the count of arguments is not same, no compatiblity possible
			return false;

		ListEnumeration enumLHS = domainListLHS.elements();
		ListEnumeration enumRHS = domainListRHS.elements();

		// Check for function subtyping!! The typeRHS should be a subtype of
		// typeLHS. This part codes the semantics of function subtyping..
		// The current semantics are
		// S:[Ds -> Rs] is a subtype of F:[Df -> Rf] iff
		//   Ds is supertype of Df
		//   Rs is a subtype of Rf
		// I have to check for RHS to be a subtype, so S === RHS, F === LHS

		// First for the domains
		while (enumLHS.hasMoreElements()) {
			// enumRHS.nextElement() has to be a supertype of 
			// enumLHS.nextElement()
			// Note the function call below: The ORDER is VERY IMPORTANT
			if (!((SOM_TypeDeclaration)enumLHS.nextElement()).
				isPredicateSubtypeOf
				((SOM_TypeDeclaration)enumRHS.nextElement()))
				// if some param doesn't fit, there is no point....
				return false;

			/*
				if (!SOM_TypeDeclaration.isLeftCompatibleWithRight
				((SOM_TypeDeclaration)enumRHS.nextElement(),
				(SOM_TypeDeclaration)enumLHS.nextElement()))
				// if some param doesn't fit, there is no point....
				return false;
				*/
		}

		// Now for the range. Range of RHS is to be a subtype of range of LHS
		return typeRHS.getRange().isPredicateSubtypeOf(typeLHS.getRange());
			/*
				SOM_TypeDeclaration.isLeftCompatibleWithRight
				(declLHS.getRange(), declRHS.getRange());
				*/
	}
}
